"""
Handles github actions like parameter matrices
"""
import ubelt as ub


def handle_yaml_grid(default, auto, arg):
    """
    Example:
        >>> default = {}
        >>> auto = {}
        >>> arg = ub.codeblock(
        >>>     '''
        >>>     matrix:
        >>>         foo: ['bar', 'baz']
        >>>     include:
        >>>         - {'foo': 'buz', 'bug': 'boop'}
        >>>     ''')
        >>> handle_yaml_grid(default, auto, arg)

        >>> default = {'baz': [1, 2, 3]}
        >>> arg = '''
        >>>     include:
        >>>     - {
        >>>       "thresh": 0.1,
        >>>       "morph_kernel": 3,
        >>>       "norm_ord": 1,
        >>>       "agg_fn": "probs",
        >>>       "thresh_hysteresis": "None",
        >>>       "moving_window_size": "None",
        >>>       "polygon_fn": "heatmaps_to_polys"
        >>>     }
        >>>     '''
        >>> handle_yaml_grid(default, auto, arg)
    """
    stdform_keys = {'matrix', 'include'}
    import ruamel.yaml
    print('arg = {}'.format(ub.urepr(arg, nl=1)))
    if arg:
        if arg is True:
            arg = 'auto'
        if isinstance(arg, str):
            if arg == 'auto':
                arg = auto
            if isinstance(arg, str):
                arg = ruamel.yaml.safe_load(arg)
    else:
        arg = {'matrix': default}
    if isinstance(arg, dict):
        arg = ub.udict(arg)
        if len(arg - stdform_keys) == 0 and (arg & stdform_keys):
            # Standard form
            ...
        else:
            # Transform matrix to standard form
            arg = {'matrix': arg}
    elif isinstance(arg, list):
        # Transform list form to standard form
        arg = {'include': arg}
    else:
        raise TypeError(type(arg))
    assert set(arg.keys()).issubset(stdform_keys)
    print('arg = {}'.format(ub.urepr(arg, nl=1)))
    basis = arg.get('matrix', {})
    if basis:
        grid = list(ub.named_product(basis))
    else:
        grid = []
    grid.extend(arg.get('include', []))
    return grid


def coerce_list_of_action_matrices(arg):
    """
    Preprocess the parameter grid input into a standard form
    """
    import ruamel.yaml
    if isinstance(arg, str):
        data = ruamel.yaml.safe_load(arg)
    else:
        data = arg.copy()
    if isinstance(data, dict):
        pass
    action_matrices = []
    if isinstance(data, list):
        for item in data:
            action_matrices.append(item)
    elif isinstance(data, dict):
        if not len(ub.udict(data) & {'matrix', 'include'}):
            data = {'matrix': data}
        action_matrices.append(data)
    return action_matrices


def prevalidate_param_grid(arg):
    """
    Determine if something may go wrong
    """

    def validate_pathlike(p):
        if isinstance(p, str):
            p = ub.Path(p)
        else:
            p = ub.Path(p)
        if p.expand().exists():
            return True
        return False

    action_matrices = coerce_list_of_action_matrices(arg)

    # TODO: this doesn't belong in a utils folder.
    src_pathlike_keys = [
        'trk.pxl.model',
        'trk.pxl.data.test_dataset',
        'crop.src',
        'act.pxl.model',
        'act.pxl.data.test_dataset',
    ]

    logs = []

    def log_issue(k, p, msg):
        logs.append((k, p, msg))
        print(f'Key {k} with {p=} {msg}')

    for item in action_matrices:
        matrix = item.get('matrix', {})
        for k in src_pathlike_keys:
            if k in matrix:
                v = matrix[k]
                v = [v] if not ub.iterable(v) else v
                for p in v:
                    if not validate_pathlike(p):
                        log_issue(k, p, 'might not be a valid path')


def expand_param_grid(arg, max_configs=None):
    """
    Our own method for specifying many combinations. Uses the github actions
    method under the hood with our own

    Ignore:
        >>> from watch.utils.util_param_grid import *  # NOQA
        >>> arg = ub.codeblock(
            '''
            - matrix:
                trk.pxl.model: [trk_a, trk_b]
                trk.pxl.data.tta_time: [0, 4]
                trk.pxl.data.set_cover_algo: [None, approx]
                trk.pxl.data.test_dataset: [D4_S2_L8]

                act.pxl.model: [act_a, act_b]
                act.pxl.data.test_dataset: [D4_WV_PD, D4_WV]
                act.pxl.data.input_space_scale: [1GSD, 4GSD]

                trk.poly.thresh: [0.17]
                act.poly.thresh: [0.13]

                exclude:
                  #
                  # The BAS A should not run with tta
                  - trk.pxl.model: trk_a
                    trk.pxl.data.tta_time: 4
                  # The BAS B should not run without tta
                  - trk.pxl.model: trk_b
                    trk.pxl.data.tta_time: 0
                  #
                  # The SC B should not run on the PD dataset when GSD is 1
                  - act.pxl.model: act_b
                    act.pxl.data.test_dataset: D4_WV_PD
                    act.pxl.data.input_space_scale: 1GSD
                  # The SC A should not run on the WV dataset when GSD is 4
                  - act.pxl.model: act_a
                    act.pxl.data.test_dataset: D4_WV
                    act.pxl.data.input_space_scale: 4GSD
                  #
                  # The The BAS A and SC B model should not run together
                  - trk.pxl.model: trk_a
                    act.pxl.model: act_b
                  # Other misc exclusions to make the output cleaner
                  - trk.pxl.model: trk_b
                    act.pxl.data.input_space_scale: 4GSD
                  - trk.pxl.data.set_cover_algo: None
                    act.pxl.data.input_space_scale: 1GSD

                include:
                  # only try the 10GSD scale for trk model A
                  - trk.pxl.model: trk_a
                    trk.pxl.data.input_space_scale: 10GSD
            ''')
        >>> grid_items = list(expand_param_grid(arg))
        >>> print('grid_items = {}'.format(ub.urepr(grid_items, nl=1, sort=0)))
        >>> from watch.utils.util_dotdict import dotdict_to_nested
        >>> print(ub.urepr([dotdict_to_nested(p) for p in grid_items], nl=-3, sort=0))
        >>> print(len(grid_items))
    """
    prevalidate_param_grid(arg)
    action_matrices = coerce_list_of_action_matrices(arg)
    num_yeilded = 0
    for item in action_matrices:
        for grid_item in extended_github_action_matrix(item):
            yield grid_item
            num_yeilded += 1
            if max_configs is not None:
                if num_yeilded >= max_configs:
                    return


def github_action_matrix(arg):
    """
    Implements the github action matrix strategy exactly as described.

    Unless I've implemented something incorrectly, I believe this method is
    limited and have extended it in :func:`extended_github_action_matrix`.

    Args:
        arg (Dict | str): a dictionary or a yaml file that resolves to a
            dictionary containing the keys "matrix", which maps parameters to a
            list of possible values. For convinieince if a single scalar value
            is detected it is converted to a list of 1 item. The matrix may
            also include an "include" and "exclude" item, which are lists of
            dictionaries that modify existing / add new matrix configurations
            or remove them. The "include" and "exclude" parameter can also be
            specified at the same level of "matrix" for convinience.

    Yields:
        item : a single entry in the grid.

    References:
        https://docs.github.com/en/actions/using-jobs/using-a-matrix-for-your-jobs#expanding-or-adding-matrix-configurations

    CommandLine:
        xdoctest -m watch.utils.util_param_grid github_action_matrix:2

    Example:
        >>> from watch.utils.util_param_grid import *  # NOQA
        >>> arg = ub.codeblock(
                 '''
                   matrix:
                     fruit: [apple, pear]
                     animal: [cat, dog]
                     include:
                       - color: green
                       - color: pink
                         animal: cat
                       - fruit: apple
                         shape: circle
                       - fruit: banana
                       - fruit: banana
                         animal: cat
                 ''')
        >>> grid_items = list(github_action_matrix(arg))
        >>> print('grid_items = {}'.format(ub.urepr(grid_items, nl=1)))
        grid_items = [
            {'fruit': 'apple', 'animal': 'cat', 'color': 'pink', 'shape': 'circle'},
            {'fruit': 'apple', 'animal': 'dog', 'color': 'green', 'shape': 'circle'},
            {'fruit': 'pear', 'animal': 'cat', 'color': 'pink'},
            {'fruit': 'pear', 'animal': 'dog', 'color': 'green'},
            {'fruit': 'banana'},
            {'fruit': 'banana', 'animal': 'cat'},
        ]

    Example:
        >>> from watch.utils.util_param_grid import *  # NOQA
        >>> arg = ub.codeblock(
                '''
                  matrix:
                    os: [macos-latest, windows-latest]
                    version: [12, 14, 16]
                    environment: [staging, production]
                    exclude:
                      - os: macos-latest
                        version: 12
                        environment: production
                      - os: windows-latest
                        version: 16
            ''')
        >>> grid_items = list(github_action_matrix(arg))
        >>> print('grid_items = {}'.format(ub.urepr(grid_items, nl=1)))
        grid_items = [
            {'os': 'macos-latest', 'version': 12, 'environment': 'staging'},
            {'os': 'macos-latest', 'version': 14, 'environment': 'staging'},
            {'os': 'macos-latest', 'version': 14, 'environment': 'production'},
            {'os': 'macos-latest', 'version': 16, 'environment': 'staging'},
            {'os': 'macos-latest', 'version': 16, 'environment': 'production'},
            {'os': 'windows-latest', 'version': 12, 'environment': 'staging'},
            {'os': 'windows-latest', 'version': 12, 'environment': 'production'},
            {'os': 'windows-latest', 'version': 14, 'environment': 'staging'},
            {'os': 'windows-latest', 'version': 14, 'environment': 'production'},
        ]

    Example:
        >>> from watch.utils.util_param_grid import *  # NOQA
        >>> arg = ub.codeblock(
                 '''
                 matrix:
                   old_variable:
                       - null
                       - auto
                 include:
                     - old_variable: null
                       new_variable: 1
                     - old_variable: null
                       new_variable: 2
                 ''')
        >>> grid_items = list(github_action_matrix(arg))
        >>> print('grid_items = {}'.format(ub.urepr(grid_items, nl=1)))
    """
    import ruamel.yaml
    if isinstance(arg, str):
        data = ruamel.yaml.safe_load(arg)
    else:
        data = arg.copy()

    matrix = data.pop('matrix', {}).copy()

    include = matrix.pop('include', data.pop('include', []))
    exclude = matrix.pop('exclude', data.pop('exclude', []))
    include = list(map(ub.udict, include))
    exclude = list(map(ub.udict, exclude))

    matrix_ = {k: (v if ub.iterable(v) else [v])
               for k, v in matrix.items()}

    orig_keys = set(matrix.keys())
    include_idx_to_nvariants = {idx: 0 for idx in range(len(include))}

    def include_modifiers(mat_item):
        """
        For each object in the include list, the key:value pairs in the object
        will be added to each of the matrix combinations if none of the
        key:value pairs overwrite any of the original matrix values. If the
        object cannot be added to any of the matrix combinations, a new matrix
        combination will be created instead. Note that the original matrix
        values will not be overwritten, but added matrix values can be
        overwritten.
        """
        grid_item = ub.udict(mat_item)
        for include_idx, include_item in enumerate(include):
            common_orig1 = (mat_item & include_item) & orig_keys
            common_orig2 = (include_item & mat_item) & orig_keys
            if common_orig1 == common_orig2:
                include_idx_to_nvariants[include_idx] += 1
                grid_item = grid_item | include_item
        return grid_item

    def is_excluded(grid_item):
        """
        An excluded configuration only has to be a partial match for it to be
        excluded. For example, the following workflow will run nine jobs: one
        job for each of the 12 configurations, minus the one excluded job that
        matches {os: macos-latest, version: 12, environment: production}, and
        the two excluded jobs that match {os: windows-latest, version: 16}.
        """
        for exclude_item in exclude:
            common1 = exclude_item & grid_item
            if common1:
                common2 = grid_item & exclude_item
                if common1 == common2 == exclude_item:
                    return True

    for mat_item in map(ub.udict, ub.named_product(matrix_)):
        grid_item = include_modifiers(mat_item)
        if not is_excluded(grid_item):
            yield grid_item

    for idx, n in include_idx_to_nvariants.items():
        if n == 0:
            grid_item = include[idx]
            yield grid_item


def extended_github_action_matrix(arg):
    """
    A variant of the github action matrix for our mlops framework that
    overcomes some of the former limitations.

    This keeps the same weird include / exclude semantics, but
    adds an additional "submatrix" component that has the following semantics.

    A submatrices is a list of dictionaries, but each dictionary may have more
    than one value, and are expanded into a list of items, similarly to a
    dictionary. In this respect the submatrix is "resolved" to a list of
    dictionary items just like "include". The difference is that when a
    common elements of a submatrix grid item matches a matrix grid item, it
    updates it with its new values and yields it immediately. Subsequent
    submatrix grid items can yield different variations of this item.
    The actions include rules are then applied on top of this.

    Args:
        arg (Dict | str): See github_action_matrix, but with new submatrices

    Yields:
        item : a single entry in the grid.

    CommandLine:
        xdoctest -m watch.utils.util_param_grid extended_github_action_matrix:2

    Example:
        >>> from watch.utils.util_param_grid import *  # NOQA
        >>> from watch.utils import util_param_grid
        >>> arg = ub.codeblock(
                 '''
                   matrix:
                     fruit: [apple, pear]
                     animal: [cat, dog]
                     submatrix:
                       - color: green
                       - color: pink
                         animal: cat
                       - fruit: apple
                         shape: circle
                       - fruit: banana
                       - fruit: banana
                         animal: cat
                 ''')
        >>> grid_items = list(extended_github_action_matrix(arg))
        >>> print('grid_items = {}'.format(ub.urepr(grid_items, nl=1)))

    Example:
        >>> from watch.utils.util_param_grid import *  # NOQA
        >>> arg = ub.codeblock(
                '''
                  matrix:
                    os: [macos-latest, windows-latest]
                    version: [12, 14, 16]
                    environment: [staging, production]
                    exclude:
                      - os: macos-latest
                        version: 12
                        environment: production
                      - os: windows-latest
                        version: 16
            ''')
        >>> grid_items = list(extended_github_action_matrix(arg))
        >>> print('grid_items = {}'.format(ub.urepr(grid_items, nl=1)))

    Example:
        >>> from watch.utils.util_param_grid import *  # NOQA
        >>> from watch.utils import util_param_grid
        >>> # Specifying an explicit list of things to run
        >>> arg = ub.codeblock(
                 '''
                 submatrices:
                    - common_variable: a
                      old_variable: a
                    - common_variable: a
                      old_variable: null
                      new_variable: 1
                    - common_variable: a
                      old_variable: null
                      new_variable: 11
                    - common_variable: a
                      old_variable: null
                      new_variable: 2
                    - common_variable: b
                      old_variable: null
                      new_variable: 22
                 ''')
        >>> grid_items = list(extended_github_action_matrix(arg))
        >>> print('grid_items = {}'.format(ub.urepr(grid_items, nl=1)))
        >>> assert len(grid_items) == 5

    Example:
        >>> from watch.utils.util_param_grid import *  # NOQA
        >>> from watch.utils import util_param_grid
        >>> arg = ub.codeblock(
                 '''
                 matrix:
                   common_variable:
                       - a
                       - b
                   old_variable:
                       - null
                       - auto
                 submatrices:
                     - old_variable: null
                       new_variable1:
                           - 1
                           - 2
                       new_variable2:
                           - 3
                           - 4
                     - old_variable: null
                       new_variable2:
                           - 33
                           - 44
                     # These wont be used because blag doesn't exist
                     - old_variable: blag
                       new_variable:
                           - 10
                           - 20
                 ''')
        >>> grid_items = list(extended_github_action_matrix(arg))
        >>> print('grid_items = {}'.format(ub.urepr(grid_items, nl=1)))
        >>> assert len(grid_items) == 14

    Example:
        >>> from watch.utils.util_param_grid import *  # NOQA
        >>> from watch.utils import util_param_grid
        >>> arg = ub.codeblock(
                 '''
                 matrix:
                   step1.src:
                       - dset1
                       - dset2
                       - dset3
                       - dset4
                   step1.resolution:
                       - 10
                       - 20
                       - 30
                 submatrices1:
                    - step1.resolution: 10
                      step2.resolution: [10, 15]
                    - step1.resolution: 20
                      step2.resolution: 20
                 submatrices2:
                    - step1.src: dset1
                      step2.src: big_dset1A
                    - step1.src: dset2
                      step2.src:
                         - big_dset2A
                         - big_dset2B
                    - step1.src: dset3
                      step2.src: big_dset3A
                 ''')
        >>> grid_items = list(extended_github_action_matrix(arg))
        >>> print('grid_items = {}'.format(ub.urepr(grid_items, nl=1)))
        >>> assert len(grid_items) == 20
    """
    import ruamel.yaml
    from watch.utils.util_yaml import Yaml
    import os
    if isinstance(arg, str):
        data = ruamel.yaml.safe_load(arg)
    else:
        data = arg.copy()

    matrix = data.pop('matrix', {}).copy()

    include = matrix.pop('include', data.pop('include', []))
    exclude = matrix.pop('exclude', data.pop('exclude', []))

    submatrices = matrix.pop('submatrices', data.pop('submatrices', []))
    submatrices = list(map(ub.udict, submatrices))

    include = list(map(ub.udict, include))
    exclude = list(map(ub.udict, exclude))

    def coerce_matrix_value(v):
        if not ub.iterable(v):
            v = [v]
        final = []
        for item in v:
            if isinstance(item, (str, os.PathLike)) and str(item).endswith(('.yaml', '.yml')):
                # use Yaml.coerce instead?
                final.extend(Yaml.load(item))
            else:
                final.append(item)
        return final

    # Special submatrices for more cartesian products, it would be good to come
    # up with a solution that does not require hard coded and a fixed number of
    # variables.
    submatrices1 = matrix.pop('submatrices1', data.pop('submatrices1', []))
    submatrices2 = matrix.pop('submatrices2', data.pop('submatrices2', []))
    submatrices3 = matrix.pop('submatrices3', data.pop('submatrices3', []))

    MULTI_SUBMATRICES = 1
    if MULTI_SUBMATRICES:
        # Try allowing for more variations. The idea is we effectively
        # want to take the cross product of multiple lists of submatrices.
        multi_submatrices = [
            submatrices, submatrices1, submatrices2, submatrices3
        ]
        multi_submatrices_ = []
        for submats in multi_submatrices:
            submats[:] = list(map(ub.udict, submats))
            submats_ = []
            for submatrix in submats:
                submatrix_ = {k: coerce_matrix_value(v)
                              for k, v in submatrix.items()}
                submats_.extend(list(map(ub.udict, ub.named_product(submatrix_))))
            multi_submatrices_.append(submats_)
    else:
        submatrices_ = []
        for submatrix in submatrices:
            submatrix_ = {k: coerce_matrix_value(v)
                          for k, v in submatrix.items()}
            submatrices_.extend(list(map(ub.udict, ub.named_product(submatrix_))))

    if len(data) != 0:
        raise Exception(f'Unexpected top level keys: {list(data.keys())}')

    matrix_ = {k: coerce_matrix_value(v)
               for k, v in matrix.items()}

    orig_keys = set(matrix.keys())
    include_idx_to_nvariants = {idx: 0 for idx in range(len(include))}

    def include_modifiers(mat_item):
        """
        For each object in the include list, the key:value pairs in the object
        will be added to each of the matrix combinations if none of the
        key:value pairs overwrite any of the original matrix values. If the
        object cannot be added to any of the matrix combinations, a new matrix
        combination will be created instead. Note that the original matrix
        values will not be overwritten, but added matrix values can be
        overwritten.
        """
        grid_item = ub.udict(mat_item)
        for include_idx, include_item in enumerate(include):
            common_orig1 = (mat_item & include_item) & orig_keys
            common_orig2 = (include_item & mat_item) & orig_keys
            if common_orig1 == common_orig2:
                include_idx_to_nvariants[include_idx] += 1
                grid_item = grid_item | include_item
        return grid_item

    def multisubmatrix_variants(mat_item, multi_submatrices_):
        # New version: every group of submatrices has the opportunity to
        # modify the item before yielding.
        curr_items = [mat_item]
        for submatrices_ in multi_submatrices_:
            curr_items = _submatrix_variants_loop(curr_items, submatrices_)
        yield from curr_items

    def _submatrix_variants_loop(mat_items, submatrices_):
        for item in mat_items:
            yield from submatrix_variants(item, submatrices_)

    def submatrix_variants(mat_item, submatrices_):
        """
        For each object in the include list, the key:value pairs in the object
        will be added to each of the matrix combinations if none of the
        key:value pairs overwrite any of the original matrix values. If the
        object cannot be added to any of the matrix combinations, a new matrix
        combination will be created instead. Note that the original matrix
        values will not be overwritten, but added matrix values can be
        overwritten.
        """
        grid_item = ub.udict(mat_item)
        any_modified = False
        for submat_item in submatrices_:
            common_orig1 = (mat_item & submat_item) & orig_keys
            common_orig2 = (submat_item & mat_item) & orig_keys
            if common_orig1 == common_orig2:
                grid_item = mat_item | submat_item
                yield grid_item
                any_modified = True
        if not any_modified:
            yield grid_item

    def is_excluded(grid_item):
        """
        An excluded configuration only has to be a partial match for it to be
        excluded. For example, the following workflow will run nine jobs: one
        job for each of the 12 configurations, minus the one excluded job that
        matches {os: macos-latest, version: 12, environment: production}, and
        the two excluded jobs that match {os: windows-latest, version: 16}.
        """
        for exclude_item in exclude:
            common1 = exclude_item & grid_item
            if common1:
                common2 = grid_item & exclude_item
                if common1 == common2 == exclude_item:
                    return True

    for mat_item in map(ub.udict, ub.named_product(matrix_)):
        if MULTI_SUBMATRICES:
            submat_gen = multisubmatrix_variants(mat_item, multi_submatrices_)
        else:
            submat_gen = submatrix_variants(mat_item, submatrices_)
        for item in submat_gen:
            item = include_modifiers(item)
            if not is_excluded(item):
                yield item

    for idx, n in include_idx_to_nvariants.items():
        if n == 0:
            grid_item = include[idx]
            yield grid_item



import os
import argparse

import torch
from tqdm import tqdm
from einops import rearrange
from torch.utils.data import DataLoader

from watch.tasks.rutgers_material_seg_v2.models import build_model
from watch.tasks.rutgers_material_seg_v2.utils.util_misc import load_cfg_file
from watch.tasks.rutgers_material_seg_v2.utils.util_image import ImageStitcher
from watch.tasks.rutgers_material_seg_v2.datasets import build_dataset, custom_collate_fn
from watch.tasks.rutgers_material_seg_v2.utils.util_misc import softmax, MATERIAL_TO_MATID


def generate_kwcoco_predictions(model, eval_loader, pred_save_path):
    device = torch.device('cuda')
    model = model.to(device)
    model = model.eval()

    # Get save dir to store prediction images.
    base_dir = '/'.join(pred_save_path.split('/')[:-1])
    exp_name = pred_save_path.split('/')[-1].split('.')[0]

    stitcher_save_dir = os.path.join(base_dir, '_material_predictions', exp_name)
    os.makedirs(stitcher_save_dir, exist_ok=True)

    pred_stitcher = ImageStitcher(stitcher_save_dir, save_backend='gdal')
    image_name_to_id = {}
    with torch.no_grad():
        for data in tqdm(eval_loader, colour='green', desc='Generating material predictions'):

            # Convert images into model format.
            # image_data = train_dataset.format_image_data(data['image'], data['buffer_mask'])
            b, c, h, w = data['image'].shape

            image_data = {}
            image_data['pixel_data'] = rearrange(data['image'], 'b c h w -> (b h w) c')
            image_data['pixel_data'] = image_data['pixel_data'].to(device)

            prediction = model.forward(image_data)

            # Resize to original crop size
            prediction = prediction.view(b, h, w, -1)
            prediction = prediction.permute(0, 3, 1, 2)

            # Convert predictions to numpy arrays.
            prediction = prediction.detach().cpu().numpy()

            # Add images to stitcher process.
            ## Get image names.
            image_names = [
                eval_loader.dataset.coco_dset.index.imgs[iid.item()]['name'] for iid in list(data['image_id'])
            ]

            # Save images from the stitcher object.
            for i in range(b):
                sm_pred = softmax(prediction[i], axis=0)
                pred_stitcher.add_image(sm_pred, image_names[i], data['crop_slice'][i], data['og_height'][i].item(),
                                        data['og_width'][i].item(), data['buffer_mask'][i])

                ## Additionally build mapping between image name and ids.
                if image_names[i] not in image_name_to_id.keys():
                    image_name_to_id[image_names[i]] = data['image_id'][i].item()

        # Finalize stitching operation and save images.
        save_paths, image_names, image_sizes = pred_stitcher.save_images()

        # Create a new kwcoco file.
        pred_coco_dset = eval_loader.dataset.coco_dset.copy()

        # Add material predictions to kwcoco file.
        for save_path, image_name, image_size in tqdm(zip(save_paths, image_names, image_sizes),
                                                      colour='green',
                                                      desc='Saving predictions'):
            # Get image ID.
            image_id = image_name_to_id[image_name]

            # Add material prediction to kwcoco file.
            pred_coco_dset.add_auxiliary_item(image_id,
                                              save_path,
                                              channels='mat_pred',
                                              width=image_size[-2],
                                              height=image_size[-1],
                                              warp_aux_to_img={'scale': 1.0})

        # Save kwcoco file.
        print('Dumping kwcoco file ... ')
        pred_coco_dset.dump(pred_save_path)
        print(f'Saved to: {pred_save_path}')

        return pred_save_path


def predict():
    parser = argparse.ArgumentParser()
    parser.add_argument('checkpoint_path', type=str, help='Path to the model weights.')
    parser.add_argument('config_path',
                        type=str,
                        help='Path to the configuration file to recreate the model and correct dataset parameters.')
    parser.add_argument('target_dataset_path',
                        type=str,
                        help='Path of the input kwcoco dataset for model to make predictions on.')
    parser.add_argument('pred_save_path', type=str, help='Where the material predictions kwcoco file will be saved.')
    parser.add_argument('--include_labels',
                        default=False,
                        action='store_true',
                        help='Include material labels in the prediction script. Will be utilized for analysis scripts.')
    args = parser.parse_args()

    # Load configuration file.
    cfg = load_cfg_file(args.config_path)

    # Generate KWCOCO file with model predictions.

    ## Get evaluation dataset and loader.
    eval_dataset = build_dataset(cfg.dataset.name, args.target_dataset_path, 'eval', **cfg.dataset.kwargs)
    eval_loader = DataLoader(eval_dataset,
                             batch_size=cfg.batch_size,
                             shuffle=False,
                             num_workers=cfg.n_workers,
                             collate_fn=custom_collate_fn)

    ## Rebuild model accroding to configurations.
    n_material_classes = len(MATERIAL_TO_MATID.keys())
    model = build_model(cfg.model.name,
                        n_in_channels=eval_dataset.n_channels,
                        n_out_channels=n_material_classes,
                        **cfg.model.kwargs)

    ## Load model with best weights.
    model = model.load_from_checkpoint(args.checkpoint_path,
                                       n_in_channels=eval_dataset.n_channels,
                                       n_out_channels=n_material_classes)

    ## Generate kwcoco file with predictions.
    _ = generate_kwcoco_predictions(model, eval_loader, args.pred_save_path)


if __name__ == '__main__':
    predict()



import kwcoco
import kwimage
import warnings
import numpy as np
import ubelt as ub
from typing import Dict, List, Any
from watch.utils.kwcoco_extensions import TrackidGenerator
from watch.utils.kwcoco_extensions import warp_annot_segmentations_to_geos
from watch.tasks.tracking.utils import TrackFunction
from watch.tasks.tracking.utils import check_only_bg
try:
    from xdev import profile
except Exception:
    profile = ub.identity


def dedupe_annots(coco_dset):
    '''
    Check for annotations with different aids that are the same geometry
    '''
    annots = coco_dset.annots()
    # NOTE: Using segmentations to dedup with segmentation data is fragile
    eq_keys = ['image_id', 'category_id', 'track_id', 'segmentation']
    eq_vals = annots.lookup(eq_keys, None)
    group_keys = [str(row_vals) for row_vals in zip(*eq_vals.values())]
    duplicates = ub.find_duplicates(group_keys)
    if duplicates:
        dup_idxs = list(ub.flatten([idxs[1:] for idxs in duplicates.values()]))
        warnings.warn(
            ub.paragraph(f'''
            There were {len(duplicates)} annotation groups
            with {dup_idxs} duplicate annotations based on
            group keys {eq_keys}
            '''))
        aids_to_remove = list(annots.take(dup_idxs))
        coco_dset.remove_annotations(aids_to_remove, verbose=1)

    return coco_dset


@profile
def remove_small_annots(coco_dset, min_area_px=1, min_geo_precision=6):
    '''
    There are several reasons for a detection to be too small to keep.
    Remove these and return the rest of the dataset.

    1. Detections that aren't well-formed polygons.
        These are simply errors.
        They show up fairly often in an arbitrary dset; TODO figure out why
        possible culprits:
            mask_to_scored_polygons?
            cropping in propagate_labels?

        >>> # xdoctest: +SKIP
        >>> d = kwcoco.CocoDataset('pred_KR_R01.kwcoco_timeagg_v1.json')
        >>> sum(are_invalid(d.annots())), d.n_annots
        6686, 13136

    2. Very small detections in pixel-space (area <1 pixel).
        These probably couldn't represent something visible,
        unless the GSD is very large.
        Skip this check by setting min_area_px=0

    3. Overly-precise geo-detections.
        Because GSD varies, and because lat-lon isn't equal-area, detections
        can be trivial in geo space but not pixel space.
        GeoJSON spec recommends a precision of 6 decimal places, which is
        ~10cm. (IARPA annotations conform to this).
        This check removes detections that are empty when rounded.
        Skip this check by setting min_geo_precision=None

    Sources:
        [1] https://pypi.org/project/geojson/#default-and-custom-precision

    Example:
        >>> import kwimage
        >>> from copy import deepcopy
        >>> from watch.tasks.tracking.normalize import remove_small_annots
        >>> from watch.demo import smart_kwcoco_demodata
        >>> coco_dset = smart_kwcoco_demodata.demo_kwcoco_with_heatmaps()
        >>> # This dset has 1 video with all images the same size
        >>> # For testing, resize one of the images so there is a meaningful
        >>> # difference between img space and vid space
        >>> scale_factor = 0.5
        >>> aff = kwimage.Affine.coerce({'scale': scale_factor})
        >>> img = coco_dset.imgs[1]
        >>> img['width'] *= scale_factor
        >>> img['height'] *= scale_factor
        >>> img['warp_img_to_vid']['scale'] = 1/scale_factor
        >>> for aux in img['auxiliary']:
        >>>     aux['warp_aux_to_img']['scale'] = np.array(
        >>>         aux['warp_aux_to_img'].get('scale', 1)) * scale_factor
        >>> annots = coco_dset.annots(gid=img['id'])
        >>> old_annots = deepcopy(annots)
        >>> dets = annots.detections.warp(aff)
        >>> # TODO this doesn't handle keypoints, and is rather brittle, is
        >>> # there a way to simply do something like:
        >>> #    annots.detections = annot.detections.warp(w)
        >>> annots.set('bbox', dets.boxes.to_coco(style='new'))
        >>> annots.set('segmentation', dets.data['segmentations'].to_coco(
        >>>     style='new'))
        >>> # test that scaling worked
        >>> assert np.all(annots.boxes.area < old_annots.boxes.area)
        >>> assert np.allclose(annots.boxes.warp(aff.inv()).area,
        >>>     old_annots.boxes.area)
        >>> # test that remove_small_annots no-ops with no threshold
        >>> # (ie there are no invalid annots here)
        >>> assert coco_dset.n_annots == remove_small_annots(deepcopy(coco_dset),
        >>>     min_area_px=0, min_geo_precision=None).n_annots
        >>> # test that annots can be removed
        >>> assert remove_small_annots(deepcopy(coco_dset), min_area_px=1e99,
        >>>     min_geo_precision=None).n_annots == 0
        >>> # test that annotations are filtered in video space
        >>> # pick a threshold above the img annot size and below the vid
        >>> # annot size; annot should not be removed
        >>> thresh = annots.boxes.area[0] + 1
        >>> assert annots.aids[0] in remove_small_annots(deepcopy(coco_dset),
        >>>     min_area_px=thresh, min_geo_precision=None).annots(
        >>>         gid=img['id']).aids
        >>> # test that some but not all annots can be removed
        >>> filtered = remove_small_annots(
        >>>     deepcopy(coco_dset), min_area_px=10000,
        >>>     min_geo_precision=None)
        >>> assert filtered.n_annots > 0 and filtered.n_annots < coco_dset.n_annots
        >>> # TODO test min_geo_precision
    '''
    if min_area_px is None:
        min_area_px = 0

    empty_aids = []
    remove_reason = []

    gid_iter = ub.ProgIter(coco_dset.index.imgs.keys(),
                           total=coco_dset.n_images,
                           desc='filter annotations')
    for gid in gid_iter:
        coco_img = coco_dset.coco_image(gid)
        annots = coco_dset.annots(gid=gid)
        for aid in annots:
            ann = coco_dset.index.anns[aid]

            if min_area_px is not None:
                pxl_sseg = coco_img._annot_segmentation(ann, space='video')

                try:
                    pxl_sseg_sh = pxl_sseg.to_multi_polygon().to_shapely()
                except ValueError:
                    remove_reason.append('invalid_shapely_conversion')
                    empty_aids.append(aid)
                    continue

                if not pxl_sseg_sh.is_valid:
                    pxl_sseg_sh = pxl_sseg_sh.buffer(0)

                if not pxl_sseg_sh.is_valid:
                    remove_reason.append('invalid_pixel_polygon')
                    empty_aids.append(aid)
                    continue

                try:
                    pxl_sseg_sh.area
                except Exception:
                    pxl_sseg_sh = pxl_sseg_sh.buffer(0)

                if pxl_sseg_sh.area <= min_area_px:
                    remove_reason.append('small pixel area {}'.format(
                        round(pxl_sseg_sh.area, 1)))
                    empty_aids.append(aid)
                    continue

            if min_geo_precision is not None:
                # TODO: could check this in UTM space instead of using rounding
                wgs_sseg = kwimage.Segmentation.coerce(
                    ann['segmentation_geos'])
                wgs_sseg_sh = wgs_sseg.to_multi_polygon().to_shapely()
                wgs_sseg_sh = shapely_round(wgs_sseg_sh,
                                            min_geo_precision).buffer(0)

                if wgs_sseg_sh.is_empty:
                    remove_reason.append('empty geos area')
                    empty_aids.append(aid)
                    continue

    if not empty_aids:
        print('Size filter: all annotations were valid')
    else:
        coco_dset.remove_annotations(empty_aids)
        keep_annots = coco_dset.annots()
        keep_tids = keep_annots.get('track_id', None)
        keep_track_lengths = ub.dict_hist(keep_tids)
        print(f'Size filter: removing {len(empty_aids)} annotations')
        print('keep_track_lengths = {}'.format(
            ub.urepr(keep_track_lengths, nl=1)))
        print(f'{len(keep_annots)=}')
        removal_reasons = ub.dict_hist(remove_reason)
        print('removal_reasons = {}'.format(ub.urepr(removal_reasons, nl=1)))

    return coco_dset


def ensure_videos(coco_dset):
    '''
    Ensure every image belongs to a video, even a dummy video
    and has a frame_index
    '''
    # HACK, TODO this is probably a kwcoco bug that needs fixed
    if 'videos' not in coco_dset.dataset:
        coco_dset.dataset['videos'] = list(coco_dset.index.videos.values())

    coco_dset.index.vidid_to_gids
    all_images = coco_dset.images()
    loose_flags = [vidid is None for vidid in all_images.get('video_id', None)]
    loose_imgs = all_images.compress(loose_flags)
    if loose_imgs:
        from watch.utils import util_time
        print(f'Warning: there are {len(loose_imgs)=} images without a video')
        # guess frame_index by date
        dt_guess = [(util_time.coerce_datetime(dc), gid)
                    for gid, dc in loose_imgs.lookup(
                        'date_captured', '1970-01-01', keepid=1).items()]
        loose_imgs = loose_imgs.take(ub.argsort(dt_guess))

        vidid = coco_dset.add_video('DEFAULT')
        for ix, gid in enumerate(loose_imgs):
            coco_dset.imgs[gid]['video_id'] = vidid
            coco_dset.imgs[gid]['frame_index'] = ix

        # This change invalidates the index, need to rebuild it.
        # TODO: add kwcoco logic to flag exactly what index needs to be rebuilt
        coco_dset._build_index()

    if __debug__:
        try:
            coco_dset.images().lookup('frame_index')
        except KeyError:
            raise AssertionError('all images in dset need a frame_index')

    return coco_dset


def dedupe_tracks(coco_dset):
    '''
    Assuming that videos are made of disjoint images, ensure that trackids
    are not shared by two tracks in different videos.
    '''
    new_trackids = TrackidGenerator(coco_dset)

    for trackid in coco_dset.index.trackid_to_aids.keys():
        annots = coco_dset.annots(track_id=trackid)

        # split each video into a separate track
        for idx, (vidid, aids) in enumerate(
                ub.group_items(
                    annots.aids,
                    coco_dset.images(annots.gids).get('video_id',
                                                      None)).items()):
            if idx > 0:
                coco_dset.annots(aids=aids).set('track_id', next(new_trackids))

    return coco_dset


def shapely_round(geom, precision):
    """
    References:
        https://gis.stackexchange.com/questions/188622
    """
    import shapely
    wkt = shapely.wkt.dumps(geom, rounding_precision=precision)
    geom2 = shapely.wkt.loads(wkt).simplify(0)
    return geom2


@profile
def normalize_phases(coco_dset,
                     use_viterbi=False,
                     t_probs=None,
                     e_probs=None,
                     baseline_keys={'salient'},
                     prediction_key='phase_transition_days'):
    '''
    Convert internal representation of phases to their IARPA standards as well
    as inserting a baseline guess for activity classification and removing
    empty tracks.

    HACK: add a Post Construction frame at the end of every track
    until we support partial sites

    The only remaining categories in the returned coco_dset should be:
        Site Preparation
        Active Construction
        Post Construction

    TODO make this a step in track_fn to take advantage of heatmap info?
    Example:
        >>> # test baseline guess
        >>> from watch.tasks.tracking.normalize import normalize_phases
        >>> from watch.tasks.tracking.normalize import normalize_phases
        >>> from watch.demo import smart_kwcoco_demodata
        >>> dset = smart_kwcoco_demodata.demo_kwcoco_with_heatmaps()
        >>> dset.remove_categories([1,3,4,5])
        >>> dset.cats[2]['name'] = 'salient'
        >>> assert dset.cats == {2: {'id': 2, 'name': 'salient'}}
        >>> # HACK, this shouldn't be needed?
        >>> # TODO file bug report
        >>> dset._build_index()
        >>> dset = normalize_phases(dset)
        >>> assert (dset.annots(track_id=1).cnames ==
        >>>     ((['Site Preparation'] * 10) +
        >>>      (['Active Construction'] * 9) +
        >>>      (['Post Construction'])))
        >>> # try again with smoothing
        >>> dset = normalize_phases(dset, use_viterbi=True)
        >>> from watch.demo import smart_kwcoco_demodata
        >>> dset = smart_kwcoco_demodata.demo_kwcoco_with_heatmaps()
        >>> dset.remove_categories([1,3,4,5])
        >>> dset.cats[2]['name'] = 'salient'
        >>> assert dset.cats == {2: {'id': 2, 'name': 'salient'}}
        >>> # HACK, this shouldn't be needed?
        >>> # TODO file bug report
        >>> dset._build_index()
        >>> dset = normalize_phases(dset)
        >>> assert (dset.annots(track_id=1).cnames ==
        >>>     ((['Site Preparation'] * 10) +
        >>>      (['Active Construction'] * 9) +
        >>>      (['Post Construction'])))
        >>> # try again with smoothing
        >>> dset = normalize_phases(dset, use_viterbi=True)
    '''
    from watch.heuristics import CATEGORIES, CNAMES_DCT, SITE_SUMMARY_CNAME
    from watch.tasks.tracking import phase
    from collections import Counter

    for cat in CATEGORIES:
        coco_dset.ensure_category(**cat)
    baseline_keys = set(baseline_keys)
    unknown_cnames = coco_dset.name_to_cat.keys() - (
        {cat['name']
         for cat in CATEGORIES} | {SITE_SUMMARY_CNAME} | baseline_keys)
    if unknown_cnames:
        print(f'removing unknown categories {unknown_cnames}')
        coco_dset.remove_categories(unknown_cnames, keep_annots=False)

    cnames_to_remove = set(
        # negative examples, no longer needed
        #CNAMES_DCT['negative']['scored'] + CNAMES_DCT['negative']['unscored'] +
        CNAMES_DCT['negative']['unscored'] +
        # should have been consumed by track_fn, TODO more robust check
        [SITE_SUMMARY_CNAME])
    coco_dset.remove_categories(cnames_to_remove, keep_annots=False)

    cnames_to_replace = (
        # 'positive'
        set(CNAMES_DCT['positive']['unscored']) |
        # 'salient' or equivalent
        baseline_keys)

    # metrics-and-test-framework/evaluation.py:1684
    cnames_to_score = set(CNAMES_DCT['positive']['scored']) | set(
        CNAMES_DCT['negative']['scored'])

    allowed_cnames = cnames_to_replace | cnames_to_score
    have_cnames = set(coco_dset.name_to_cat)
    if not have_cnames.issubset(allowed_cnames):
        raise AssertionError(
            ub.paragraph(f'''
            Unhandled Class Names
            {allowed_cnames=}
            {have_cnames=}
            unknown={have_cnames - allowed_cnames}
            '''))

    # Transform phase labels of each track
    #

    log = Counter()
    for trackid, annot_ids in coco_dset.index.trackid_to_aids.items():
        n_anns = len(annot_ids)
        if n_anns > 1:
            annots = coco_dset.annots(track_id=trackid)
            has_missing_labels = bool(set(annots.cnames) - cnames_to_score)
            has_good_labels = bool(set(annots.cnames) - cnames_to_replace)
            if has_missing_labels and has_good_labels:
                # if we have partial coverage, interpolate from good labels
                log.update(['partial class labels'])
                coco_dset = phase.interpolate(coco_dset, trackid, cnames_to_keep=cnames_to_score)
            else:
                if has_missing_labels:
                    # else, predict site prep for the first half of the track
                    # and then active construction for the second half with
                    # post construction on the last frame
                    log.update(['no class labels'])
                    coco_dset = phase.baseline(coco_dset, trackid)
                else:
                    log.update(['full class labels'])
    print('label status of tracks: ', log)

    #
    # Continue transforming phase labels, now with smoothing and deduping
    #

    annots = coco_dset.annots()
    old_cnames_dct = dict(zip(annots.aids, annots.cnames))

    for trackid, annot_ids in coco_dset.index.trackid_to_aids.items():
        n_anns = len(annot_ids)
        annots = coco_dset.annots(track_id=trackid)

        if n_anns > 1:

            if use_viterbi:
                # with xdev.embed_on_exception_context():
                smoothed_cnames = phase.class_label_smoothing(
                    annots.cnames,
                    transition_probs=t_probs,
                    emission_probs=e_probs
                )
                annots.set('category_id', [
                    coco_dset.name_to_cat[name]['id']
                    for name in smoothed_cnames
                ])

            # after viterbi, the sequence of phases is in the correct order

            # If the track ends before the end of the video, and the last frame
            # is not post construction, add another frame of post construction
            # coco_dset = phase.dedupe_background_anns(coco_dset, trackid)
            coco_dset = phase.ensure_post(coco_dset, trackid)

        annots = coco_dset.annots(track_id=trackid)
        is_empty = check_only_bg(annots.cnames)
        EMPTY_TRACK_BEHAVIOR = 'ignore'

        if is_empty:
            print(
                f'apply {EMPTY_TRACK_BEHAVIOR} to {trackid=} with cats {set(annots.cnames)}'
            )
            if EMPTY_TRACK_BEHAVIOR == 'delete':
                coco_dset.remove_annotations(annots.aids)
            elif EMPTY_TRACK_BEHAVIOR == 'flag':
                annots.set('status', 'system_rejected')
            elif EMPTY_TRACK_BEHAVIOR == 'revert':
                annots.cnames = list(
                    ub.dict_subset(old_cnames_dct, annots.aids).values())
            elif EMPTY_TRACK_BEHAVIOR == 'ignore':
                pass
            else:
                raise ValueError(EMPTY_TRACK_BEHAVIOR)

    #
    # Phase prediction - do it all at once for efficiency
    # TODO do this before viterbi so we can use this as an input in the future
    #

    ann_field = 'phase_transition_days'

    # exclude untracked annots which might be unrelated
    annots = coco_dset.annots(
        list(ub.flatten(coco_dset.index.trackid_to_aids.values())))

    if len(annots) > 0:
        has_prediction_heatmaps = all(
            kwcoco.FusedChannelSpec.coerce(prediction_key).as_set().issubset(
                coco_dset.coco_image(gid).channels.fuse().as_set())
            for gid in set(annots.gids))
        if has_prediction_heatmaps:
            phase_transition_days = phase.phase_prediction_heatmap(
                annots, coco_dset, prediction_key)
            annots.set(ann_field, phase_transition_days)
        else:
            for trackid in coco_dset.index.trackid_to_aids.keys():
                _annots = coco_dset.annots(track_id=trackid)
                phase_transition_days = phase.phase_prediction_baseline(_annots)
                _annots.set(ann_field, phase_transition_days)

        #
        # Fixup phase prediction
        #

        # TODO do something with transition preds for phases which were altered
        FIXUP_TRANSITION_PRED = 0
        if FIXUP_TRANSITION_PRED:
            n_diff_annots = sum(
                np.array(annots.cnames) == np.array(old_cnames_dct.values()))
            if n_diff_annots > 0:
                raise NotImplementedError

    return coco_dset


def normalize_sensors(coco_dset):
    '''
    Convert internal representations of sensors to their IARPA standards
    '''
    # FIXME: should pull from heuristics
    from watch.heuristics import TE_SENSOR_NAMES
    sensor_dict = TE_SENSOR_NAMES
    good_sensors = set(sensor_dict.values())
    for img in coco_dset.imgs.values():
        try:
            sensor = img['sensor_coarse']
            if sensor not in good_sensors:
                img['sensor_coarse'] = sensor_dict[sensor]
        except KeyError:
            sensor = img.get('sensor_coarse', None)
            # name = img.get('name', img['file_name'])
            import warnings
            warnings.warn(
                f'image has unknown sensor {sensor} in tag={coco_dset.tag}')

    return coco_dset


def dedupe_dates(coco_dset):
    '''
    Ensure a tracked kwcoco file has at most 1 annot per track per date. [1]

    There are several potential ways to do this.
     - take highest-resolution sensor [currently done]
     - take image with best coverage (least nodata)
     - take latest time
     - majority-vote labels/average scores
     - average heatmaps before polygons are created

    Given that this probably has a minimal impact on scores, the safest method
    is chosen.

    References:
        [1] https://smartgitlab.com/TE/metrics-and-test-framework/-/issues/63

    Example:
        >>> from watch.tasks.tracking.normalize import *  # NOQA
        >>> import watch
        >>> import kwarray
        >>> coco_dset = watch.coerce_kwcoco('watch-msi', geodata=True, dates=True)
        >>> # Add 0-4 duplicate images to each video
        >>> rng = kwarray.ensure_rng(613544)
        >>> gids_to_duplicate = list(ub.flatten([rng.choice(gs, rng.randint(0, 4)) for gs in coco_dset.videos().images]))
        >>> for gid in gids_to_duplicate:
        >>>     img1 = ub.udict(coco_dset.index.imgs[gid]) - {'id'}
        >>>     img1['name'] = img1['name'] + '_duplicated'
        >>>     coco_dset.add_image(**img1)
        >>> coco_dset_with_dups = coco_dset.copy()
        >>> coco_dset_fixed = dedupe_dates(coco_dset.copy())
        >>> assert coco_dset_fixed.n_images < coco_dset_with_dups.n_images
    '''
    from watch.utils import util_time
    sensor_priority = {
        'WorldView': 6,
        'WorldView 1': 5,
        'Planet': 4,
        'Sentinel-2': 3,
        'Landsat 8': 2,
        'Landsat 7': 1
    }
    for trackid in coco_dset.index.trackid_to_aids.keys():
        annots = coco_dset.annots(track_id=trackid)
        dates = [util_time.coerce_datetime(d).date() for d in annots.images.lookup('date_captured')]
        fixs = annots.images.lookup('frame_index')

        is_sorted = lambda arr: np.all(arr[:-1] <= arr[1:])  # noqa
        if not all(date_frame_sorted := (
                    is_sorted(dates),
                    is_sorted(fixs))):
            # this should never print
            print(f'WARNING: {trackid=} {date_frame_sorted=}')

    # remove full images instead of iterating over tracks for efficiency
    # not possible for some other removal methods, but it is for this one
    gids_to_remove = []
    for vidid in coco_dset.index.vidid_to_gids.keys():
        images = coco_dset.images(vidid=vidid)
        dates = [util_time.coerce_datetime(d).date() for d in images.lookup('date_captured')]
        dup_dates_to_idxs: Dict[Any, List[int]] = ub.find_duplicates(dates)
        # If we have any duplicates for a day, lookup their priorities and
        # remove all but the one with the highest priority.
        for date, dup_idxs in dup_dates_to_idxs.items():
            dup_images = images.take(dup_idxs)
            dup_sensors = dup_images.lookup('sensor_coarse')
            dup_priorities = [
                sensor_priority.get(s, -1) for s in dup_sensors]
            keep_idx = ub.argmax(dup_priorities)
            remove_gids = list(set(dup_images) - {dup_images[keep_idx]})
            print(f'removing {len(remove_gids)} dup imgs from {date} in {vidid=}')
            gids_to_remove.extend(remove_gids)

    # coco_dset.remove_annotations(aids_to_remove, verbose=1)
    coco_dset.remove_images(gids_to_remove, verbose=2)
    return coco_dset


@profile
def normalize(
        coco_dset,
        track_fn,
        gt_dset=None,
        viz_out_dir=None,
        use_viterbi=False,
        # t_probs=None,  # for viterbi
        # e_probs=None,  # for viterbi
        **track_kwargs):
    '''
    Driver function to apply all normalizations

    TODO:
        Rename this to something like run_tracker. This is the entry point to
        the main tracking pipeline.

    Example:
        >>> import kwcoco as kc
        >>> from watch.tasks.tracking.normalize import *
        >>> from watch.tasks.tracking.from_polygon import OverlapTrack
        >>> # create demodata
        >>> d = kc.CocoDataset.demo()
        >>> ann_dct = d.anns[1]
        >>> d.remove_annotations(range(1,12))
        >>> ann_dct.pop('keypoints')
        >>> ann_dct.pop('id')
        >>> for i in range(1,4):
        >>>     ann_dct.update(image_id=i)
        >>>     d.add_annotation(**ann_dct)
        >>> for img, sensor in zip(d.imgs.values(), ['WV', 'S2', 'L8']):
        >>>     img['sensor_coarse'] = sensor
        >>> d.remove_categories(range(2,9))
        >>> d.cats[1]['supercategory'] = None
        >>> d.cats[1]['name'] = 'change'
        >>> d.images().set('channels', 'rgb')
        >>> # test everything except geo-info
        >>> def _normalize_annots(coco_dset):
        >>>     coco_dset = dedupe_annots(coco_dset)
        >>>     coco_dset = remove_small_annots(coco_dset,
        >>>         min_geo_precision=None)
        >>>     return coco_dset
        >>> coco_dset = d.copy()
        >>> coco_dset = _normalize_annots(coco_dset)
        >>> assert coco_dset.anns == d.anns
        >>> coco_dset = ensure_videos(coco_dset)
        >>> assert coco_dset.index.vidid_to_gids[1] == coco_dset.imgs.keys()
        >>> n_existing_annots = coco_dset.n_annots
        >>> coco_dset = OverlapTrack().apply_per_video(coco_dset)
        >>> assert set(coco_dset.annots().get('track_id')) == {1}
        >>> assert coco_dset.n_annots == n_existing_annots
        >>> coco_dset = dedupe_tracks(coco_dset)
        >>> assert set(coco_dset.annots().get('track_id')) == {1}
        >>> coco_dset = normalize_phases(coco_dset, baseline_keys={'change'})
        >>> assert (coco_dset.annots().cnames ==
        >>> ['Site Preparation', 'Site Preparation', 'Post Construction'])
        >>> coco_dset = normalize_sensors(coco_dset)
        >>> assert (coco_dset.images().get('sensor_coarse') ==
        >>>     ['WorldView', 'Sentinel-2', 'Landsat 8'])
    '''

    DEBUG_JSON_SERIALIZABLE = 0
    if DEBUG_JSON_SERIALIZABLE:
        from watch.utils.util_json import debug_json_unserializable

    if viz_out_dir is not None:
        try:
            viz_out_dir = ub.Path(viz_out_dir)
        except TypeError:
            viz_out_dir = ub.Path('_assets/tracking_visualization')
            print(f'setting default {viz_out_dir=}')
        viz_out_dir.mkdir(parents=True, exist_ok=True)

    def _normalize_annots(coco_dset):
        print(f'coco_dset.n_anns={coco_dset.n_annots}')
        coco_dset = dedupe_annots(coco_dset)
        warp_annot_segmentations_to_geos(coco_dset)
        coco_dset = remove_small_annots(coco_dset,
                                        min_area_px=0,
                                        min_geo_precision=None)
        print(f'coco_dset.n_anns={coco_dset.n_annots}')
        # coco_dset._build_index()

        return coco_dset

    if len(coco_dset.anns) > 0:
        coco_dset = _normalize_annots(coco_dset)
    coco_dset = ensure_videos(coco_dset)

    if DEBUG_JSON_SERIALIZABLE:
        debug_json_unserializable(coco_dset.dataset, 'Input to normalize: ')

    # apply tracks
    assert issubclass(track_fn, TrackFunction), 'must supply a valid track_fn!'

    # fixup the track kwargs when they come in via json
    for k, v in track_kwargs.items():
        if isinstance(v, str) and v.lower() == 'none':
            track_kwargs[k] = None

    if DEBUG_JSON_SERIALIZABLE:
        debug_json_unserializable(coco_dset.dataset, 'Before apply_per_video: ')

    tracker: TrackFunction = track_fn(**track_kwargs, viz_out_dir=viz_out_dir)
    print('track_kwargs = {}'.format(ub.urepr(track_kwargs, nl=1)))
    print('{} {}'.format(tracker.__class__.__name__, ub.urepr(tracker.__dict__, nl=1)))
    out_dset = tracker.apply_per_video(coco_dset)

    if DEBUG_JSON_SERIALIZABLE:
        debug_json_unserializable(out_dset.dataset, 'After apply_per_video: ')

    # normalize and add geo segmentations
    out_dset = _normalize_annots(out_dset)
    out_dset._build_index()
    print('After normalizing: track ids',
          set(out_dset.annots().get('track_id', None)))

    out_dset = dedupe_tracks(out_dset)

    if viz_out_dir is not None:
        from watch.tasks.tracking.visualize import viz_track_scores
        out_pth = viz_out_dir / 'track_scores.jpg'
        viz_track_scores(out_dset, out_pth, gt_dset)

    if isinstance(use_viterbi, str):
        parts = use_viterbi.split(',')
        assert len(parts) == 2
        t_probs, e_probs = parts
    else:
        t_probs = 'default'
        e_probs = 'default'

    phase_kw = dict(
        use_viterbi=use_viterbi,
        t_probs=t_probs,
        e_probs=e_probs
    )
    if 'key' in track_kwargs:  # assume this is a baseline (saliency) key
        k = track_kwargs['key']
        if ub.iterable(k):
            k = set(k)
        else:
            k = {k}
        phase_kw['baseline_keys'] = k
    out_dset = normalize_phases(out_dset, **phase_kw)

    if DEBUG_JSON_SERIALIZABLE:
        debug_json_unserializable(out_dset.dataset, 'After normalize_phases: ')

    out_dset = normalize_sensors(out_dset)

    # HACK, ensure out_dset.index is up to date
    out_dset._build_index()

    out_dset = dedupe_dates(out_dset)

    if DEBUG_JSON_SERIALIZABLE:
        debug_json_unserializable(out_dset.dataset, 'Output of normalize: ')

    if viz_out_dir is not None:
        # visualize predicted sites with true sites

        # TODO think more about key handling
        from watch.tasks.tracking.visualize import visualize_videos
        from watch.tasks.tracking.utils import _validate_keys
        from dataclasses import asdict
        fg, bg = _validate_keys(
            asdict(tracker).get('key', None),
            asdict(tracker).get('bg_key', None))
        keys = '|'.join([*fg, *bg])
        visualize_videos(out_dset, viz_out_dir / 'gif', keys, gt_dset)

    return out_dset


# Note: smooth transition while changing the name of "normalize"
# Make a long search replacable name that we can fix later.
run_tracking_pipeline = normalize



"""
See Old Version:
    ../../../scripts/run_uky_invariants_for_baseline.py

SeeAlso:
    ~/code/watch-smartflow-dags/KIT_TA2_PREEVAL10_PYENV.py
"""
import os
import subprocess
import ubelt as ub
import scriptconfig as scfg

from watch.cli.baseline_framework_kwcoco_egress import baseline_framework_kwcoco_egress  # noqa: 501
from watch.cli.baseline_framework_kwcoco_ingress import baseline_framework_kwcoco_ingress  # noqa: 501


class TeamFeatLandcoverConfig(scfg.DataConfig):
    """
    Run DZYNE landcover feature computation as baseline framework component
    """
    input_path = scfg.Value(None, type=str, position=1, required=True, help=ub.paragraph(
            '''
            Path to input T&E Baseline Framework JSON
            '''))
    input_region_path = scfg.Value(None, type=str, position=2, required=True, help=ub.paragraph(
            '''
            Path to input T&E Baseline Framework Region definition JSON
            '''))
    output_path = scfg.Value(None, type=str, position=3, required=True, help='S3 path for output JSON')
    model_path = scfg.Value(None, type=str, required=True, help='File path to landcover model')
    aws_profile = scfg.Value(None, type=str, help=ub.paragraph(
            '''
            AWS Profile to use for AWS S3 CLI commands
            '''))
    dryrun = scfg.Value(False, isflag=True, short_alias=['d'], help='Run AWS CLI commands with --dryrun flag')
    outbucket = scfg.Value(None, type=str, required=True, short_alias=['o'], help=ub.paragraph(
            '''
            S3 Output directory for STAC item / asset egress
            '''))
    requester_pays = scfg.Value(False, isflag=True, short_alias=['r'], help=ub.paragraph(
            '''
            Run AWS CLI commands with `--requestor_payer requester` flag
            '''))
    newline = scfg.Value(False, isflag=True, short_alias=['n'], help=ub.paragraph(
            '''
            Output as simple newline separated STAC items
            '''))
    jobs = scfg.Value(1, type=int, short_alias=['j'], help='Number of jobs to run in parallel')


def main():
    config = TeamFeatLandcoverConfig.cli(strict=True)
    print('config = {}'.format(ub.urepr(dict(config), nl=1, align=':')))
    run_landcover_for_baseline(**config)


def run_landcover_for_baseline(input_path,
                               input_region_path,
                               output_path,
                               model_path,
                               outbucket,
                               aws_profile=None,
                               dryrun=False,
                               requester_pays=False,
                               newline=False,
                               jobs=1):
    from watch.utils.util_framework import download_region
    if aws_profile is not None:
        aws_base_command =\
            ['aws', 's3', '--profile', aws_profile, 'cp']
    else:
        aws_base_command = ['aws', 's3', 'cp']

    if dryrun:
        aws_base_command.append('--dryrun')

    # 1. Ingress data
    print("* Running baseline framework kwcoco ingress *")
    ingress_dir = '/tmp/ingress'
    ingress_kwcoco_path = baseline_framework_kwcoco_ingress(
        input_path,
        ingress_dir,
        aws_profile,
        dryrun)

    # 2. Download and prune region file
    print("* Downloading and pruning region file *")
    local_region_path = '/tmp/region.json'
    download_region(
        input_region_path=input_region_path,
        output_region_path=local_region_path,
        aws_profile=aws_profile,
        strip_nonregions=True,
    )

    # 2. Generate Landcover features
    print("* Generating landcover features *")
    dzyne_landcover_features_kwcoco_path = os.path.join(
        ingress_dir, 'dzyne_landcover_kwcoco.json')

    subprocess.run(['python', '-m', 'watch.tasks.landcover.predict',
                    '--dataset', ingress_kwcoco_path,
                    '--deployed', model_path,
                    '--output', dzyne_landcover_features_kwcoco_path,
                    '--num_workers', '2',
                    '--with_hidden', '32',
                    '--select_images', '.sensor_coarse == "S2"',
                    '--device', '0'],
                   check=True)

    # 3. Combining landcover features with input features to pass to BAS
    print("* Combining input features with computed landcover features *")
    combo_features_kwcoco_path = os.path.join(
        ingress_dir, 'features_combo_with_landcover_kwcoco.json')
    subprocess.run(['python', '-m', 'watch.cli.coco_combine_features',
                    '--src',
                    ingress_kwcoco_path,
                    dzyne_landcover_features_kwcoco_path,
                    '--dst',
                    combo_features_kwcoco_path],
                   check=True)

    # 4. Egress (envelop KWCOCO dataset in a STAC item and egress;
    #    will need to recursive copy the kwcoco output directory up to
    #    S3 bucket)
    print("* Egressing KWCOCO dataset and associated STAC item *")
    baseline_framework_kwcoco_egress(combo_features_kwcoco_path,
                                     local_region_path,
                                     output_path,
                                     outbucket,
                                     aws_profile=None,
                                     dryrun=False,
                                     newline=False)


if __name__ == "__main__":
    main()



import watch
import json
import geojson
import ubelt as ub
import dateutil
import datetime
from collections import defaultdict
# import pandas as pd
# import geopandas as gpd


def default_feat():
    return {
        'is_occluded': 'False',
        'is_site_boundary': 'True',
        'type': 'observation',
        'source': 'DUMMY',
        'sensor_name': 'Sentinel-2',
        'score': 1.0
    }


def to_date(date_str):
    return dateutil.parser.parse(date_str).date()


# TODO should phase_transition_days be in a list []? It is in the delivered
# system.
def extend(site_summary, site):

    sstart_date = site_summary['properties']['start_date']
    send_date = site_summary['properties']['end_date']
    ssd, sed = to_date(sstart_date), to_date(send_date)
    start_date = site['features'][0]['properties']['start_date']
    end_date = site['features'][0]['properties']['end_date']
    sd, ed = to_date(start_date), to_date(end_date)
    site_id = site_summary['properties']['site_id']

    if ssd < sd:
        print(f'extending start of {site_id} from {sd} to {ssd}')
        geom = site['features'][1]['geometry']
        days = site['features'][1]['properties']['misc_info'][
            'phase_transition_days']
        days[0] += (sd - ssd).days
        obs0_feat = {}
        obs0_feat['geometry'] = geom
        obs0_feat['properties'] = default_feat()
        obs0_feat['properties']['current_phase'] = 'Site Preparation'
        obs0_feat['properties']['observation_date'] = start_date
        obs0_feat['properties']['misc_info'] = {
            'phase_transition_days': [days]
        }
        site['features'].insert(1, obs0_feat)
        site['features'][0]['properties']['start_date'] = sstart_date

    if ed < sed:
        print(f'extending end of {site_id} from {ed} to {sed}')
        try:
            phase = site['features'][-2]['properties']
        except IndexError:
            phase = 'Active Construction'
        site['features'][-1]['properties']['current_phase'] = phase
        if phase == 'Post Construction':
            days = 1.0
        else:
            days = (sed - ed).days
        site['features'][-1]['properties']['misc_info'][
            'phase_transition_days'] = [days]
        obsn_feat = {}
        obsn_feat['geometry'] = site['features'][-1]['geometry']
        obsn_feat['properties'] = default_feat()
        obsn_feat['properties']['current_phase'] = 'Post Construction'
        obsn_feat['properties']['observation_date'] = send_date
        obsn_feat['properties']['misc_info'] = {'phase_transition_days': [1.0]}
        site['features'].append(obsn_feat)
        site['features'][0]['properties']['end_date'] = send_date

    return site


def extend_clone(bas_site, site):
    '''
    TODO does not handle phase_transition_days or enforce temporal consistency
    in phases.
    '''

    sstart_date = bas_site['features'][0]['properties']['start_date']
    send_date = bas_site['features'][0]['properties']['end_date']
    ssd, sed = to_date(sstart_date), to_date(send_date)
    start_date = site['features'][0]['properties']['start_date']
    end_date = site['features'][0]['properties']['end_date']
    sd, ed = to_date(start_date), to_date(end_date)
    site_id = bas_site['features'][0]['properties']['site_id']

    n_existing_feats = len(site['features']) - 1

    site_feats = defaultdict(list)
    for f in site['features'][1:]:
        k = to_date(f['properties']['observation_date'])
        site_feats[k].append(f)

    bas_site_feats = defaultdict(list)
    for f in bas_site['features'][1:]:
        k = to_date(f['properties']['observation_date'])
        bas_site_feats[k].append(f)

    if ssd < sd:
        print(f'extending start of {site_id} from {sd} to {ssd}')
        site['features'][0]['properties']['start_date'] = sstart_date

    if ed < sed:
        print(f'extending end of {site_id} from {ed} to {sed}')
        site['features'][0]['properties']['end_date'] = send_date

    for k in bas_site_feats:
        if k not in site_feats:
            site_feats[k] = bas_site_feats[k]

    site['features'] = (site['features'][:1] +
                        list(ub.flatten(ub.sorted_keys(site_feats).values())))

    n_after_feats = len(site['features']) - 1
    print(f'added {n_after_feats - n_existing_feats} features to {site_id}')

    return site


def create(site_summary):
    geom = site_summary['geometry']
    start_date = site_summary['properties']['start_date']
    end_date = site_summary['properties']['end_date']
    site_id = site_summary['properties']['site_id']

    sd, ed = to_date(start_date), to_date(end_date)
    tot_days = (ed - sd).days
    mid_days = int(tot_days / 2)
    end_days = tot_days - mid_days
    mid_date = (sd + datetime.timedelta(days=mid_days)).isoformat()
    print(f'creating {site_id} from {sd} to {ed}')

    site = []

    site_feat = site_summary.copy()
    site_feat['properties']['type'] = 'site'
    site_feat['properties']['region_id'] = '_'.join(site_id.split('_')[:-1])
    site_feat['properties']['misc_info'] = {}
    site.append(site_feat)

    obs0_feat = {}
    obs0_feat['geometry'] = geom
    obs0_feat['properties'] = default_feat()
    obs0_feat['properties']['current_phase'] = 'Site Preparation'
    obs0_feat['properties']['observation_date'] = start_date
    obs0_feat['properties']['misc_info'] = {
        'phase_transition_days': [mid_days]
    }
    site.append(obs0_feat)

    obs1_feat = {}
    obs1_feat['geometry'] = geom
    obs1_feat['properties'] = default_feat()
    obs1_feat['properties']['current_phase'] = 'Active Construction'
    obs0_feat['properties']['observation_date'] = mid_date
    obs1_feat['properties']['misc_info'] = {
        'phase_transition_days': [end_days]
    }
    site.append(obs1_feat)

    obs2_feat = {}
    obs2_feat['geometry'] = geom
    obs2_feat['properties'] = default_feat()
    obs2_feat['properties']['current_phase'] = 'Post Construction'
    obs0_feat['properties']['observation_date'] = end_date
    obs2_feat['properties']['misc_info'] = {'phase_transition_days': [1.0]}
    site.append(obs2_feat)

    return {'type': 'FeatureCollection', 'features': site}


def main(region_models_pth,
         site_models_pth,
         out_region_models_pth,
         out_site_models_pth,
         modify_region_models_pth,
         modify_site_models_pth=None,
         create_missing=True,
         extend_existing=True):
    '''
    Modifies site models based on site summaries, and saves new site models and
    site summaries.

    Args:
        modify_site_models_pth: If given, create/extend from site models
        instead of site summaries
        create_missing: Create a new site model for any site summary that
        doesn't have one
        extend_existing: Extend the temporal duration of any site model smaller
        than its site summary

    Ignore:
        >>> from watch.cli.extend_sc_sites import main
        >>> main(
        >>>     'sc_out_region_models',
        >>>     'sc_out_site_models',
        >>>     'extended_sc_out_region_models',
        >>>     'extended_sc_out_site_models',
        >>>     'cropped_region_models_bas',
        >>> )
        >>> import jsonschema
        >>> import json
        >>> import watch
        >>> import ubelt as ub
        >>> REGION_SCHEMA = watch.rc.load_region_model_schema()
        >>> region_pths = ub.Path('extended_sc_out_region_models').glob('*.geojson')
        >>> regions = [json.load(open(p)) for p in region_pths]
        >>> for region in regions:
        >>>     # jsonschema.validate(region, schema=REGION_SCHEMA)
        >>>     pass
        >>> SITE_SCHEMA = watch.rc.load_site_model_schema()
        >>> site_pths = ub.Path('extended_sc_out_site_models').glob('*.geojson')
        >>> sites = [json.load(open(p)) for p in site_pths]
        >>> for site in sites:
        >>>     # jsonschema.validate(site, schema=SITE_SCHEMA)
        >>>     pass
        >>> # new way to validate sites
        >>> from iarpa_smart_metrics.evaluation import SiteStack
        >>> for p in site_pths:
        >>>     s = SiteStack(p)

    '''
    mr_pths = watch.utils.util_gis.coerce_geojson_paths(
        modify_region_models_pth)
    rm_pths = watch.utils.util_gis.coerce_geojson_paths(region_models_pth)
    ms_pths = watch.utils.util_gis.coerce_geojson_paths(modify_site_models_pth)
    sm_pths = watch.utils.util_gis.coerce_geojson_paths(site_models_pth)
    assert len(mr_pths) == 1, 'need 1 region -> n site models'
    assert len(rm_pths) == 1, 'need 1 region -> n site models'
    if len(sm_pths) < 1:
        print('warning: no existing site models found')

    msite_summaries = {}
    for pth in mr_pths:
        rm = json.load(open(pth))
        for f in rm['features']:
            if (f['properties']['type'] == 'site_summary'
                    and f['properties']['model_content'] == 'proposed'):
                msite_summaries[f['properties']['site_id']] = f

    site_summaries = {}
    others = []
    for pth in rm_pths:
        rm = json.load(open(pth))
        for f in rm['features']:
            if (f['properties']['type'] == 'site_summary'
                    and f['properties']['model_content'] == 'proposed'):
                site_summaries[f['properties']['site_id']] = f
            else:
                others.append(f)

    msite_models = {}
    for pth in ms_pths:
        sm = json.load(open(pth))
        f = sm['features'][0]
        if (f['properties']['type'] == 'site'
                and f['properties']['model_content'] == 'proposed'):
            msite_models[f['properties']['site_id']] = sm

    site_models = {}
    for pth in sm_pths:
        sm = json.load(open(pth))
        f = sm['features'][0]
        if (f['properties']['type'] == 'site'
                and f['properties']['model_content'] == 'proposed'):
            site_models[f['properties']['site_id']] = sm

    default_summary_settings = next(iter(site_summaries.values())).copy()
    for k in ['start_date', 'end_date', 'site_id']:
        default_summary_settings['properties'].pop(k)
    rid = others[0]['properties']['region_id']

    if extend_existing:
        for k in site_models:
            if k in msite_models:
                sm = extend_clone(msite_models[k], site_models[k])
            else:
                sm = extend(msite_summaries[k], site_models[k])
            site_models[k] = sm
            ss = site_summaries[k]
            ss['start_date'] = sm['features'][0]['properties']['start_date']
            ss['end_date'] = sm['features'][0]['properties']['end_date']
            site_summaries[k] = ss

    if create_missing:
        for k in msite_summaries:
            if k not in site_models:
                if k in msite_models:
                    site_models[k] = msite_models[k]
                else:
                    sm = create(msite_summaries[k])
                ss = default_summary_settings.copy()
                ss['site_id'] = k
                ss['start_date'] = sm['features'][0]['properties'][
                    'start_date']
                ss['end_date'] = sm['features'][0]['properties']['end_date']
                site_models[k] = sm
                site_summaries[k] = ss

    new_rm = geojson.FeatureCollection(others + list(site_summaries.values()))
    with open(
            ub.Path(out_region_models_pth).ensuredir() / f'{rid}.geojson',
            'w') as f:
        json.dump(new_rm, f, indent=2)

    for sm in site_models.values():
        sid = sm['features'][0]['properties']['site_id']
        with open(
                ub.Path(out_site_models_pth).ensuredir() / f'{sid}.geojson',
                'w') as f:
            json.dump(sm, f, indent=2)
